#include <stdio.h>

#include "sage++user.h"

#include "anls_att.h"

extern char * mixstring;

void add_secdad(SgStatement * s, SgArrayRefExp * arrayexp, int counter)
{	
  SgSymbol * dadsec = s->getSymbolInCurrentScope(mixstring, "sdd", NULL, counter);
  SgExpression * lhs = new SgExpression (VAR_REF, NULL, NULL, dadsec);
  
  SgSymbol * newsec = s->getSymbolInCurrentScope(mixstring, "new_array_section", NULL);
  SgFunctionRefExp * rhs = new SgFunctionRefExp(*newsec);
  
  int dim =0;
  for (int i=0; i<arrayexp->numberOfSubscripts(); i++) 
    if (!arrayexp->isScalarOnDim(i))
      dim++;
  
  SgValueExp * dimexp = new SgValueExp(dim);
  rhs->addArg(*dimexp);
  
  SgSymbol * dad = s->getSymbolInCurrentScope(mixstring, "dad", arrayexp->symbol()->identifier());
  SgExpression * dadexp = new SgExpression (VAR_REF, NULL, NULL, dad);
  rhs->addArg(*dadexp);
  
  SgAssignStmt * assign = new SgAssignStmt(*lhs, *rhs);
  
  s->insertStmtBefore(*assign, *s->controlParent());

}

void delete_secdad(SgStatement * s, SgArrayRefExp * arrayexp, int counter)
{
  SgSymbol * deletesec= s->getSymbolInCurrentScope(mixstring, "delete_array", NULL);
  SgCallStmt * funcall = new SgCallStmt(*deletesec);
  SgSymbol * dadsec = s->getSymbolInCurrentScope(mixstring, "sdd", NULL,counter);
  SgExpression * dadsecexp = new SgExpression (VAR_REF, NULL, NULL, dadsec);
  funcall->addArg(*dadsecexp);
  
  s->insertStmtAfter(*funcall, *s->controlParent());
}

void define_secdad(SgStatement * s, SgArrayRefExp *arrayexp, int counter)
{
  for (int i=0; i<arrayexp->numberOfSubscripts(); i++) {
    
    int dimcounter=0;

    if (!arrayexp->isScalarOnDim(i)) {
      SgSymbol * symb= s->getSymbolInCurrentScope(mixstring, "set_array_triplet", NULL);
      SgCallStmt * funcall= new SgCallStmt(*symb);

      {
        SgSymbol * symb = s->getSymbolInCurrentScope(mixstring, "sdd", NULL,counter);
        SgExpression * dadsecexp = new SgExpression(VAR_REF, NULL, NULL, symb);

        funcall->addArg(*dadsecexp);
      }
      {
        dimcounter++;
        SgValueExp * secdim = new SgValueExp (dimcounter);
        funcall->addArg(*secdim);
      }
      {
	SgTripletExp * triplet=(SgTripletExp *)(arrayexp->subscript(i));
	for (int j=0; j<triplet->numberOfItems(); j++) {
	  SgExpression * item = triplet->item(j);
	  if (item->isInteger()) {
	    SgValueExp * itemvalue = new SgValueExp (item->valueInteger());
	    funcall->addArg(*itemvalue);
	  }
	  else {
	    funcall->addArg(*item);
	  }
	  if (triplet->numberOfItems()<3) {
	    SgValueExp * defaultstride = new SgValueExp(1);
	    funcall->addArg(*defaultstride);
	  }
	}
      }
      {
        SgSymbol * symb =s->getSymbolInCurrentScope(mixstring, "dad", arrayexp->symbol()->identifier());
        SgExpression * dadexp = new SgExpression(VAR_REF, NULL, NULL, symb);

        funcall->addArg(*dadexp);
      }
      {
        SgValueExp * dim = new SgValueExp (i+1);
        funcall->addArg(*dim);
      }

      s->insertStmtBefore(*funcall, *s->controlParent());
    }
    else {
      SgSymbol * symb= s->getSymbolInCurrentScope(mixstring, "set_array_scalar", NULL);
      SgCallStmt * funcall= new SgCallStmt(*symb);
      {
        SgSymbol * symb = s->getSymbolInCurrentScope(mixstring, "sdd", NULL, counter);
        SgExpression * dadsecexp = new SgExpression(VAR_REF, NULL, NULL, symb);

        funcall->addArg(*dadsecexp);
      }
      {
        SgExpression * scalar = arrayexp ->subscript(i);
        funcall->addArg(*scalar);
      }
      {
        SgSymbol * symb =s->getSymbolInCurrentScope(mixstring, "dad", arrayexp->symbol()->identifier());
        SgExpression * dadexp = new SgExpression(VAR_REF, NULL, NULL, symb);

        funcall->addArg(*dadexp);
      };

      {
        SgValueExp * dim = new SgValueExp (i+1);
        funcall->addArg(*dim);
      }
      s->insertStmtBefore(*funcall, *s->controlParent());
    }
  }
};

void change_procstmt(SgStatement *s)
{
  SgCallStmt * call = (SgCallStmt *)s;

  SgCallStmt * newcall = new SgCallStmt(*call->name());

  for (int i=0; i<call->numberOfArgs(); i++) {

    SgExpression * argexp = call->arg(i);

    switch(argexp->variant()) {
    case VAR_REF: {
      SgSymbol * argname = call->arg(i)->symbol();
      SgExpression * var = new SgExpression (VAR_REF, NULL, NULL, argname);
      newcall->addArg(*var);
      break;
      };

    case ARRAY_REF: {
      SgArrayRefExp * arrayexp = (SgArrayRefExp *) argexp;
      SgSymbol * argname = s->getSymbolInCurrentScope(argexp->symbol()->identifier());
      
      SgArrayRefExp * var = new SgArrayRefExp(*argname);
      
      SgExpression * address = NULL;
      
      if (argexp->symbol()->numberOfAttributes(Align_Def)!=0) {
	// we need the first address of the template section;
	
	int hasScalarDim=0;
	
	for (int j=0; j<arrayexp->numberOfSubscripts(); j++) {
	  if (arrayexp->subscript(j)->variant()!=DDOT) {
	    hasScalarDim=1;
	    break;
	  };
	}
	
	if (!hasScalarDim) {
	  address=new SgValueExp(0);
	}
	else {
	  for (int dimI = 0; dimI<arrayexp->numberOfSubscripts(); dimI++) {

	    SgExpression * multi=NULL;

	    if (arrayexp->subscript(dimI)->variant()!=DDOT) {
	      SgSymbol * loc = s->getSymbolInCurrentScope(mixstring, "loc", NULL);
	      SgFunctionRefExp * locfun = new SgFunctionRefExp (*loc);
	      {
		SgSymbol * rng = s->getSymbolInCurrentScope(mixstring, "rng", NULL);
		SgFunctionRefExp * rngfun = new SgFunctionRefExp (*rng);
		
		SgSymbol * dad = s->getSymbolInCurrentScope(mixstring, "dad", argexp->symbol()->identifier());
		SgExpression * dadexp = new SgExpression (VAR_REF, NULL, NULL, dad);
		
		rngfun->addArg(*dadexp);
		
		SgValueExp * dim = new SgValueExp(dimI+1);
		
		rngfun->addArg(*dim);
		
		locfun->addArg(*rngfun);
	      }

	      {
		locfun->addArg(*arrayexp->subscript(dimI));
	      }

	      multi=locfun;
	      
	      SgSymbol * str = s->getSymbolInCurrentScope(mixstring, "str", NULL);
	      SgFunctionRefExp * strfun = new SgFunctionRefExp (*str);
		
	      SgSymbol * dad = s->getSymbolInCurrentScope(mixstring, "dad", argexp->symbol()->identifier());
	      SgExpression * dadexp = new SgExpression (VAR_REF, NULL, NULL, dad);
	      
	      strfun->addArg(*dadexp);
	      
	      SgValueExp * dim = new SgValueExp(dimI+1);
	      strfun->addArg(*dim);
	      
	      (*multi)= (*multi)*(*strfun);

	      address = address + (*multi);
	    }
	  }
	}
	var->addSubscript(*address);
      }
      else {
	printf("not finished!\n");
      }
      
      newcall->addArg(*var);
      break;
    };
    
    default:
      printf("Not touched exp in call statement!\n");
      break;
    };

    SgSymbol * dadsymb=s->getSymbolInCurrentScope(mixstring, "sdd", NULL, i+1);
    SgExpression * dadexp = new SgExpression (VAR_REF, NULL, NULL, dadsymb);
    newcall->addArg(*dadexp);
  }
  s->insertStmtBefore(*newcall, *s->controlParent());
}

void trans_procstmt(SgStatement * s) 
{
  SgCallStmt * call = (SgCallStmt *)s;

  int arraycounter=0;

  for (int i=0; i<call->numberOfArgs(); i++) {
    //printf("The id is %d\n", call->arg(i)->id());
    SgExpression * var = call->arg(i);
    switch (var->variant()) {
    case VAR_REF:
      
      break;

    case ARRAY_REF:
      {
	arraycounter++;
	SgArrayRefExp * arrayexp = (SgArrayRefExp *) var;
	add_secdad(s, arrayexp, arraycounter);
	delete_secdad(s, arrayexp, arraycounter);
	define_secdad(s, arrayexp, arraycounter);
      }
      break;

    default:
      printf("Expression in call not touched!\n");
      break;

    };
  };

  change_procstmt(s);

};

//--------------------;

void change_prochedr(SgStatement * s)
{
  SgProcHedrStmt *  prochedr = (SgProcHedrStmt *) s;
  int org_numberOfParameters = prochedr->numberOfParameters();

  SgSymbol * parameters[10];

  int i;

  for (i=0; i<org_numberOfParameters; i++) 
    parameters[i]=prochedr->parameter(i);

  for (i=0; i<org_numberOfParameters; i++) {
    SgSymbol * parameter=parameters[i];
    SgSymbol * dad = s->getSymbolInCurrentScope(mixstring, "dad", parameter->identifier());
    prochedr->insertParameterAfter(*parameter, *dad);

  }
}



